# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/00_core.ipynb 6
from . import polytope
from . import inner_point
from . import walk
from . import polytope_examples




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_polytope.ipynb.

# %% auto 0
__all__ = ['Polytope', 'NFDPolytope']

# %% ../nbs/01_polytope.ipynb 2
import torch
import numpy as np

from typing import  Union
import gc
import warnings

from fastcore.basics import patch

# %% ../nbs/01_polytope.ipynb 3
from . import warningss

# %% ../nbs/01_polytope.ipynb 4
class Polytope:
    
    def __init__(self, 
                A_in:Union[torch.Tensor, np.ndarray], # Inequality Restriction Matrix 
                b_in:Union[torch.Tensor, np.ndarray], # Inequality Vector
                dtype=torch.float16, # torch dtype
                device = None, # device used cpu or cuda
                copy:bool=False, # bool for creating a copy of the restrictions
                requires_grad:bool=False # If we want to keep the gradient
                 ) -> None:
        
        self.dtype = dtype
        if device is None:
            self.device = 'cpu' if A_in.get_device()==-1 else f'cuda:{A_in.get_device()}'
        else:
            self.device = device
        self._assert_valid_device_(device=device)
        
        self._check_dtype_()
        self._check_intra_constraint_dimensions_(A_in,b_in,'Inequality')
        
        self.copy = copy
        if copy:
            warnings.warn('The object will create a copy of the tensors, so memory usage will increase')
        else:
            warnings.warn('The object will not create a copy of the tensors, so modifications will be reflected in the object')

        if requires_grad:
            warnings.warn('The tensors will accumalate the gradient of the operations')
        self.requires_grad = requires_grad
        
        self.mI = A_in.shape[0]
        self.n = A_in.shape[1]        
        self.A_in = self._process_tensor_or_array_(A_in,'A_in')
        self.b_in = self._process_tensor_or_array_(b_in,'b_in')
        gc.collect()
    
    
    def _check_intra_constraint_dimensions_(self, 
                        A:Union[torch.Tensor, np.ndarray], 
                        b:Union[torch.Tensor, np.ndarray], 
                        constraint:str=None):
    
        assert(A.shape[0] == b.shape[0]), f"{constraint} has dimension mismatch: A has shape {A.shape} and b {b.shape}."
    
    
    def _process_tensor_or_array_(self, 
                                 A:Union[torch.Tensor, np.ndarray], 
                                 restriction:str=None):
        gc.collect()
        dtype = self.dtype
        if isinstance(A, torch.Tensor): # Torch Tensor
            # A is already a PyTorch tensor
            if self.copy:
                tensor_A = A.clone()
            else:
                tensor_A = A
            if dtype is not None and A.dtype != dtype:
                tensor_A = tensor_A.to(dtype)
            return tensor_A.requires_grad_(self.requires_grad)
        elif isinstance(A, np.ndarray): # Numpy array
            # A is a NumPy array, convert it to a PyTorch tensor
            if self.copy:
                tensor_A = torch.tensor(A, dtype=dtype)
            else:
                tensor_A = torch.from_numpy(A, dtype=dtype)
            return tensor_A.requires_grad_(self.requires_grad)
        else:
            raise ValueError(f"Input {restriction} must be a NumPy array or PyTorch tensor")

    def _check_dtype_(self):
        valid_dtypes = (torch.float16, torch.float32, torch.float64)
        assert self.dtype in valid_dtypes, f'{self.dtype} is not a valid PyTorch float data type. Valid dtypes: {valid_dtypes}'
        if '16' in str(self.dtype): 
            long_message = f'The dtype {self.dtype} is typically used with GPU architectures. If you are using CPU, consider using 32 or 64-bit dtypes. \
Certain operations may be casted to 32 or 64 bits to enhance numerical stability.'

            warnings.warn(long_message)
            
    def _assert_valid_device_(self, device):
        """
        Asserts that the given device is valid in PyTorch.

        Args:
        device (str): The device to check, e.g., 'cuda:0', 'cpu'.
        """
        # Check if the device is 'cpu' or 'cuda'. If 'cuda', check if it's available
        if device == 'cpu' or (device.startswith('cuda') and torch.cuda.is_available()):
            # If device is cuda, additionally check if the specified index is valid
            if device.startswith('cuda'):
                device_index = int(device.split(':')[1]) if ':' in device else 0
                assert device_index < torch.cuda.device_count(), "CUDA device index is out of range."
        else:
            raise AssertionError(f"Device '{device}' is not valid. Please choose 'cpu' or a valid 'cuda' device.")


# %% ../nbs/01_polytope.ipynb 5
@patch
def send_to_device(self:Polytope, device:str=None):
    self.A_in = self.A_in.to(device)
    self.b_in = self.b_in.to(device)
    self.device = device
    
@patch
def cast_precision(self:Polytope, dtype=None):
    if dtype is None:
        dtype = self.dtype
    self.A_in = self.A_in.to(dtype)
    self.b_in = self.b_in.to(dtype)
    self.dtype = dtype 

# %% ../nbs/01_polytope.ipynb 6
@patch
def __str__(self:Polytope):
    string = f'Numeric Precision (dtype) {self.dtype}\n'
    string = string + f'Device: {self.device}\n'
    string = string + f'A_in: {self.A_in.shape} \n'
    string = string + f'b_in: {self.b_in.shape}'
    return string

@patch
def __repr__(self:Polytope):
    return self.__str__()

# %% ../nbs/01_polytope.ipynb 7
class NFDPolytope(Polytope):
    
    def __init__(self, 
                A_in:Union[torch.Tensor, np.ndarray], 
                b_in:Union[torch.Tensor, np.ndarray], 
                A_eq:Union[torch.Tensor, np.ndarray], 
                b_eq:Union[torch.Tensor, np.ndarray], 
                dtype=torch.float32, # Default dtype for Non-Fully-Dimensioonal Polytopes is 32 bits
                device = None,
                copy:bool=False,
                requires_grad:bool=False
                 ) -> None:
        
        self._check_intra_constraint_dimensions_(A_in,b_in,'Inequality')
        self._check_intra_constraint_dimensions_(A_eq,b_eq,'Equality')
        self._check_inter_constraint_dimensions_(A_in, A_eq)
        
        if '16' in str(dtype):
            warnings.warn('Float16 precision was selected, since there are some equality restrictions a Projection matrix must be computed. Be sure to evaluate the numerical stability of the algorithm. We suggest using float32 or bigger.')

        super().__init__(A_in, b_in, dtype, device,copy, requires_grad)
        self.mE = A_eq.shape[0]
        self.A_eq = self._process_tensor_or_array_(A_eq,'A_eq')
        self.b_eq = self._process_tensor_or_array_(b_eq,'b_eq')
        self.projection_matrix = None
        gc.collect()
        
        
    def _check_inter_constraint_dimensions_(self,A_in,A_eq):
        
        assert(A_in.shape[1] == A_eq.shape[1]), f"Constraints have dimension mismatch: A_in has shape {A_in.shape} and  A_eq {A_eq.shape}."
        
        

# %% ../nbs/01_polytope.ipynb 8
@patch
def send_to_device(self:NFDPolytope, device:str):
    Polytope.send_to_device(self,device)
    self.device = None
    self.A_eq = self.A_eq.to(device)
    self.b_eq = self.b_eq.to(device)
    if self.projection_matrix is not None:
        self.projection_matrix = self.projection_matrix.to(device)
    self.device = device

@patch
def cast_precision(self:NFDPolytope, dtype=None):
    Polytope.cast_precision(self, dtype)
    self.dtype = None
    self.A_eq = self.A_eq.to(dtype)
    self.b_eq = self.b_eq.to(dtype)
    if self.projection_matrix is not None:
        self.projection_matrix = self.projection_matrix.to(dtype)
    self.dtype = dtype 

# %% ../nbs/01_polytope.ipynb 9
@patch
def compute_projection_matrix(self:NFDPolytope, device:str, solver_precision=torch.float64):

    ## Set Device
    original_device = self.device
    self.send_to_device(device)
    
    ## Set Precision
    if '64' in str(solver_precision):
        solver_precision_tensor = torch.float64
    elif '32' in str(solver_precision):
        solver_precision_tensor = torch.float32
    elif '16' in str(solver_precision):
        solver_precision_tensor = torch.float16
    original_precision = self.dtype
    error_precision = self.dtype
        
    if ('cuda' not in device) & ('16' in str(original_precision)):
        warnings.warn('Float16 precision was chosen for the polytope, but the "device=cpu" option is selected. Tensors will be temporarily cast to float32 for computing the Projection Matrix and stability evaluation. If you wish to use float16 precision, please select "device=cuda".')
        solver_precision_tensor = torch.float32
        error_precision = torch.float32
    self.cast_precision(solver_precision_tensor)
    
    # Compute (A A')^(-1)
    A_eq_t = torch.transpose(self.A_eq, 0, 1)
    A_eq_mm_A_eq_t = torch.matmul(self.A_eq, A_eq_t)
    
    # Compute (A A')^(-1)A
    la = torch.linalg.solve(A_eq_mm_A_eq_t,self.A_eq)

    # Check numerical stability of (A A')^(-1) (AA') - I
    est = torch.mm(la.to(error_precision), A_eq_t.to(error_precision))
    est = torch.max(torch.abs(est - torch.eye(est.shape[0], device=device).to(error_precision)))
    print(f"Max non zero error for term (A A')^(-1)A at precision {error_precision}: ", est)
    del est

    # Compute I - A'(A A')^(-1)A
    projection_matrix = torch.matmul(A_eq_t,la)
    projection_matrix = torch.eye(projection_matrix.shape[0], device=device) - projection_matrix

    # Free Memory
    del A_eq_mm_A_eq_t
    del A_eq_t
    del la
    gc.collect()
    self.projection_matrix =  projection_matrix
        
    self.send_to_device(original_device)    
    self.cast_precision(original_precision)
              
        

# %% ../nbs/01_polytope.ipynb 10
@patch
def project_h(self:NFDPolytope, x):
    assert(self.projection_matrix is not None), f'Projection Matrix has not been computed.'
        
    x_projected = torch.matmul(self.projection_matrix,x)
    return x_projected

# %% ../nbs/01_polytope.ipynb 11
@patch
def __str__(self:NFDPolytope):
    string = Polytope.__str__(self)
    string = string + f'\nA_eq: {self.A_eq.shape} \nb_eq: {self.b_eq.shape}'
    if self.projection_matrix is not None:
        string = string + f'\nProjection Matrix: {self.projection_matrix.shape}'
    return string

@patch
def __repr__(self:NFDPolytope):
    return self.__str__()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/b_utils.ipynb.

# %% auto 0
__all__ = ['print_progress']

# %% ../nbs/b_utils.ipynb 2
def print_progress(iteration, total, prefix='', length=30, fill='█'):
    percent = ("{0:.1f}").format(100 * (iteration / float(total)))
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + '-' * (length - filled_length)
    return f'\r{prefix} |{bar}| {percent}%'





# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_polytope_examples.ipynb.

# %% auto 0
__all__ = ['Hypercube', 'Simplex']

# %% ../nbs/04_polytope_examples.ipynb 2
import torch

from .polytope import Polytope, NFDPolytope
from .inner_point import ChebyshevCenter

# %% ../nbs/04_polytope_examples.ipynb 3
class Hypercube(Polytope):
    
    def __init__(self,
                 n:int=None,
                 dtype=torch.float16, 
                 device=None,
                 copy: bool = False, 
                 requires_grad: bool = False) -> None:
        device = 'cpu' if device is None else device
        A_in = torch.cat((torch.eye(n), torch.eye(n) * -1.0), dim=0).to(dtype)
        b_in = torch.ones(2 * n, dtype=dtype).view(-1, 1)        
        super().__init__(A_in, b_in, dtype,device, copy, requires_grad)

# %% ../nbs/04_polytope_examples.ipynb 5
class Simplex(NFDPolytope):
    
    def __init__(self,
                 n:int=None,
                 dtype=torch.float16, 
                 device=None,
                 copy: bool = False, 
                 requires_grad: bool = False) -> None:
        device = 'cpu' if device is None else device
        A_in = torch.eye(n).to(dtype) * -1.0
        b_in = torch.empty(n, 1, dtype=dtype)
        b_in.fill_(0.0)

        # Create Equalities
        A_eq = torch.empty(1, n, dtype=dtype)
        A_eq.fill_(1.0)
        b_eq = torch.empty(1, 1, dtype=dtype)
        b_eq.fill_(1.0) 
             
        super().__init__(A_in, b_in, A_eq, b_eq,dtype, device,copy, requires_grad)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/a_warnings.ipynb.

# %% auto 0
__all__ = ['custom_warning_format', 'custom_warning']

# %% ../nbs/a_warnings.ipynb 2
import warnings
import textwrap

# %% ../nbs/a_warnings.ipynb 3
# Custom warning format function with line wrapping and indentation
def custom_warning_format(message, category, filename, lineno, line=None):
    formatted_message = textwrap.fill(message, width=100, initial_indent='  ', subsequent_indent='  ')
    return f"{filename}:{lineno}: {category.__name__}:\n{formatted_message}\n"

# Custom warning handler
def custom_warning(message, category, filename, lineno, file=None, line=None):
    formatted_message = custom_warning_format(str(message), category, filename, lineno)
    print(formatted_message)

warnings.showwarning = custom_warning
warnings.filterwarnings("always")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_walk.ipynb.

# %% auto 0
__all__ = ['sample_inner_points', 'create_h', 'draw_uniform', 'update_progress_bars', 'check_numerical_stability', 'walk']

# %% ../nbs/03_walk.ipynb 2
import torch

from typing import Union
from tqdm import tqdm

from .polytope import Polytope, NFDPolytope
from .utils import print_progress

# %% ../nbs/03_walk.ipynb 3
def sample_inner_points(X0, z):
    n, k = X0.shape
    if z > k:
        # Pad by repeating vectors from X to fill z columns
        num_repeats = z // k
        remainder = z % k
        I = torch.cat([X0] * num_repeats + [X0[:, :remainder]], dim=1)
    elif z < k:
        # Take the first z columns from X
        I = X0[:, :z]
    else:
        # z == k, I equals X
        I = X0.clone()
    return I

# %% ../nbs/03_walk.ipynb 4
def create_h(n,z,generator,dtype,device):
    """
    Creates a Tensor (z x n x 1) where each entry ~ N(0,1). Automatically detects the
    precision 64 bits or 32 bits.
    -------------
    :param n:   int
                Dimension of the Space where the Polytope Lives
    :param z:   int
                Padding Parameter
    :param generator:
    :param device:  String, default = cpu
                    Hardware used to make the computations and allocate the result.
                    If equal to cpu then the CPUs are used for computing the inverse.
                    If equal to cuda then the a GPU is used for computing the inverse.
    -------------
    :return:    Torch Tensor
                Tensor (z x n x 1) where each entry ~ N(0,1)Contains a tensor

    """
    h = torch.randn((n, z), dtype=dtype, device=device, generator=generator)

    return h

# %% ../nbs/03_walk.ipynb 5
def draw_uniform(z, generator,dtype,device='cpu'):
    """
    Creates a tensor (z x 1) where each entry ~ U(0,1). Automatically detects the
    precision 64 bits or 32 bits.
    -------------
    :param n:   int
                Dimension of the Space where the Polytope Lives
    :param z:   int
                Padding Parameter
    :param generator:
    :param device:  String, default = cpu
                    Hardware used to make the computations and allocate the result.
                    If equal to cpu then the CPUs are used for computing the inverse.
                    If equal to cuda then the a GPU is used for computing the inverse.
    -------------
    :return:    Torch Tensor
                Tensor (z x 1) where each entry ~ U(0,1)

    """
    u = torch.rand((1, z), dtype=dtype, device=device, generator=generator)

    return u
    

# %% ../nbs/03_walk.ipynb 6
import sys

def update_progress_bars(burned, thinning, t, T, new_t: bool = False):
    # Generate progress bar string for burned samples
    if thinning !=0:
        progress_burned = print_progress(iteration=burned, prefix='% of burned samples', total=thinning, length=30)

    # Clear the current line
    sys.stdout.write('\033[K')

    if new_t:
        # Generate progress bar string for iid samples
        progress_iid = print_progress(iteration=t, prefix='% of iid samples', total=T, length=30)
        # Print both progress bars
        if thinning !=0:
            print(progress_burned, end='\n')
        print(progress_iid, end='\n')
        # Move cursor up two lines for the next update
        #sys.stdout.write("\033[F\033[F")
    else:
        # Print only the progress bar for burned samples
        if thinning !=0:
            print(progress_burned, end='\r')

    # If new_t is False, the cursor will remain on the same line ready for the next update


# %% ../nbs/03_walk.ipynb 7
def check_numerical_stability(tensor,tensor_name:str, dtype):
        # Check for NaN values
        has_nan = torch.isnan(tensor).any()
        # Check for infinite values
        has_inf = torch.isinf(tensor).any()
        if (has_nan) | (has_inf):
            print(f"{tensor_name} has NaN values, consider increasing your current precision {dtype}")
            print(f"{tensor_name} has Infinite value consider increasing your current precision {dtype}")
            return True
        return False

# %% ../nbs/03_walk.ipynb 8
def walk(
        polytope:Union[Polytope, NFDPolytope], # Polytope Object
        X0:torch.Tensor,    # Initial Interior point(s) of dim=(n,k).     
                            #- If z > k (number of columns in X0), pad  by repeating vectors from X to fill z columns.
                            #- If z < k, take the first z columns from X0.
                            #- If z == k, equal to X0.
        z:int=1, # The number of simultaneous to be executed (padding parameter).
        T:int=1, # id-iterations, total_iid_points = T*z. Each iid iteration will burn the samples established by the thinning factor.
        warm:int=None, # Number of iid-iterations needed to warm. The walk will execute warm steps before saving the points.
        thinning:int=None, # Thinning Factor. Default O(n^3)
        device:str = None, # Deveice to use, cpu or cuda
        seed:int=None, # Seed for Pseudo-Random Number Generation
        verbosity:int=1, # Verbosity of the execution
        ) -> torch.Tensor:
        
        
    ## Check validity 
    # Device
    device = device if device is not None else polytope.device
    assert(device in ['cpu', 'cuda']), print('The device is not correctly specified: ', device,
                                        '\n Please choose cpu or cuda')
    # X0 dimension
    assert(X0.shape[0] == polytope.n)
    
    ## Set min and max values
    min_ = torch.finfo(polytope.dtype).min + 2
    max_ = torch.finfo(polytope.dtype).max - 2
    eps =  torch.finfo(polytope.dtype).eps
    tolerance = eps*3
    if verbosity > 1:
        print(f'Minimum number allowed {min_}')
        print(f'Maximum number allowed {max_}')
        print(f'Eps: ', eps)
        print(f'Values close to zero will be converted to 3eps or -3eps:', tolerance)
        
    
    
    ## Set seed
    random_gen = torch.Generator(device=device)
    if seed:
        random_gen.manual_seed(seed)
    else:
        random_gen.seed()
    
    
    ## Check Dimensions
    n = polytope.n
    mI = polytope.mI
    if isinstance(polytope, NFDPolytope) :
        mE = polytope.mE
    else:
        mE=None            
    if verbosity >=1:
        print('n: ', n, '  mI:', mI, '  mE:', mE, '  z:', z)
        
    
    ## Compute/set thinning factor
    if thinning is not None:
        pass
    else:
        thinning = int(n * n * n)
        if verbosity >= 1:
            print('Automatic Thinning factor: ', thinning)
            
    if warm is not None:
        pass
    else:
        warm = thinning
        if verbosity >= 1:
            print('Warming iterations: ', thinning)
     
            
    ## Prepare and send Matrices
    init_x0 = sample_inner_points(X0,z).to(device)  
    polytope.send_to_device(device)
    
    ## Iteration Loop
    t = 1
    burned = 0
    dtype = polytope.dtype
    x = init_x0.to(dtype)
    X = torch.empty((T,n, z), dtype=dtype)
    
    while t <= T:
        
        h = create_h(n, z, generator=random_gen, dtype=dtype,device=device)        
        if isinstance(polytope, NFDPolytope) :
            d = polytope.project_h(h)
        else:
            d = h
        ## Compute B - AX / AD
        numerator = polytope.b_in - torch.matmul(polytope.A_in, x)
        denominator = torch.matmul(polytope.A_in, d).to(dtype)
        denominator = torch.where((denominator >= 0) & (denominator < tolerance), torch.full_like(denominator, tolerance), denominator)
        denominator = torch.where((denominator < 0) & (denominator > -tolerance), torch.full_like(denominator, -tolerance), denominator)
        check = check_numerical_stability(denominator,'Denominator (A_I)D ', dtype)
        if check:
            return denominator
        
        check = check_numerical_stability(numerator,'Numerator b_I - A_I(X) ', dtype)
        if check:
            return numerator
        
        # Overwrite numerator to keep memory free
        frac = numerator / denominator
        #frac = torch.clamp(frac, min=max_, max=max_)

        check = check_numerical_stability(frac,'Fraction [b_I - A_I(X)]/[(A_I)D] ', dtype)
        if check:
            return frac, numerator, denominator
        
        # From the positive denominators you want the smallest one
        lambda_pos = torch.min((denominator < 0.0).to(dtype) * max_ + # For negative denominator use max val
                               (denominator > 0.0).to(dtype) * frac, 0).values.to(dtype)[None,:]
        # From the negative denominators you want the biggest
        lambda_neg = torch.max((denominator > 0.0).to(dtype) * min_ +
                               (denominator < 0.0).to(dtype) * frac, 0).values.to(dtype)[None,:]
        
    
        # Uniform draw
        u = draw_uniform(z, generator=random_gen, dtype=dtype,device=polytope.device)
        theta = (1.0 - u) * lambda_pos + u * lambda_neg

        # New X
        x = x.to(dtype) + d.to(dtype) * theta.to(dtype)
        
        if verbosity >=1:
            update_progress_bars(burned, thinning, t, T)        
        
        if (warm > 0) & (burned >= thinning):
            warm = warm - 1
            burned = 0
        
        # Manage the burning rate and save points
        if burned >= thinning:
            X[t-1,:, :] = x.to('cpu')        
            
            
            if verbosity >=1:
                update_progress_bars(burned, thinning, t, T, new_t=True)
            t = t + 1
            burned = 0

        burned = burned + 1

    return X

    



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'released',
                'doc_baseurl': '/mhar',
                'doc_host': 'https://sonder-art.github.io',
                'git_url': 'https://github.com/sonder-art/mhar',
                'lib_path': 'mhar'},
  'syms': { 'mhar.core': {},
            'mhar.inner_point': {'mhar.inner_point.ChebyshevCenter': ('inner_point.html#chebyshevcenter', 'mhar/inner_point.py')},
            'mhar.polytope': { 'mhar.polytope.NFDPolytope': ('polytope.html#nfdpolytope', 'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope.__init__': ('polytope.html#nfdpolytope.__init__', 'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope.__repr__': ('polytope.html#nfdpolytope.__repr__', 'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope.__str__': ('polytope.html#nfdpolytope.__str__', 'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope._check_inter_constraint_dimensions_': ( 'polytope.html#nfdpolytope._check_inter_constraint_dimensions_',
                                                                                                  'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope.cast_precision': ('polytope.html#nfdpolytope.cast_precision', 'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope.compute_projection_matrix': ( 'polytope.html#nfdpolytope.compute_projection_matrix',
                                                                                        'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope.project_h': ('polytope.html#nfdpolytope.project_h', 'mhar/polytope.py'),
                               'mhar.polytope.NFDPolytope.send_to_device': ('polytope.html#nfdpolytope.send_to_device', 'mhar/polytope.py'),
                               'mhar.polytope.Polytope': ('polytope.html#polytope', 'mhar/polytope.py'),
                               'mhar.polytope.Polytope.__init__': ('polytope.html#polytope.__init__', 'mhar/polytope.py'),
                               'mhar.polytope.Polytope.__repr__': ('polytope.html#polytope.__repr__', 'mhar/polytope.py'),
                               'mhar.polytope.Polytope.__str__': ('polytope.html#polytope.__str__', 'mhar/polytope.py'),
                               'mhar.polytope.Polytope._assert_valid_device_': ( 'polytope.html#polytope._assert_valid_device_',
                                                                                 'mhar/polytope.py'),
                               'mhar.polytope.Polytope._check_dtype_': ('polytope.html#polytope._check_dtype_', 'mhar/polytope.py'),
                               'mhar.polytope.Polytope._check_intra_constraint_dimensions_': ( 'polytope.html#polytope._check_intra_constraint_dimensions_',
                                                                                               'mhar/polytope.py'),
                               'mhar.polytope.Polytope._process_tensor_or_array_': ( 'polytope.html#polytope._process_tensor_or_array_',
                                                                                     'mhar/polytope.py'),
                               'mhar.polytope.Polytope.cast_precision': ('polytope.html#polytope.cast_precision', 'mhar/polytope.py'),
                               'mhar.polytope.Polytope.send_to_device': ('polytope.html#polytope.send_to_device', 'mhar/polytope.py')},
            'mhar.polytope_examples': { 'mhar.polytope_examples.Hypercube': ( 'polytope_examples.html#hypercube',
                                                                              'mhar/polytope_examples.py'),
                                        'mhar.polytope_examples.Hypercube.__init__': ( 'polytope_examples.html#hypercube.__init__',
                                                                                       'mhar/polytope_examples.py'),
                                        'mhar.polytope_examples.Simplex': ('polytope_examples.html#simplex', 'mhar/polytope_examples.py'),
                                        'mhar.polytope_examples.Simplex.__init__': ( 'polytope_examples.html#simplex.__init__',
                                                                                     'mhar/polytope_examples.py')},
            'mhar.utils': {'mhar.utils.print_progress': ('b_utils.html#print_progress', 'mhar/utils.py')},
            'mhar.walk': { 'mhar.walk.check_numerical_stability': ('walk.html#check_numerical_stability', 'mhar/walk.py'),
                           'mhar.walk.create_h': ('walk.html#create_h', 'mhar/walk.py'),
                           'mhar.walk.draw_uniform': ('walk.html#draw_uniform', 'mhar/walk.py'),
                           'mhar.walk.sample_inner_points': ('walk.html#sample_inner_points', 'mhar/walk.py'),
                           'mhar.walk.update_progress_bars': ('walk.html#update_progress_bars', 'mhar/walk.py'),
                           'mhar.walk.walk': ('walk.html#walk', 'mhar/walk.py')},
            'mhar.warningss': { 'mhar.warningss.custom_warning': ('a_warnings.html#custom_warning', 'mhar/warningss.py'),
                                'mhar.warningss.custom_warning_format': ('a_warnings.html#custom_warning_format', 'mhar/warningss.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_inner_point.ipynb.

# %% auto 0
__all__ = ['ChebyshevCenter']

# %% ../nbs/02_inner_point.ipynb 2
from scipy.optimize import linprog
import numpy as np
import torch

from typing import Union
import warnings
import gc

from .polytope import Polytope, NFDPolytope
from . import warningss

# %% ../nbs/02_inner_point.ipynb 3
def ChebyshevCenter(
                    polytope:Union[Polytope, NFDPolytope], 
                    lb=None, # Lowerbound (lb <= x ), if unknown leave it as None 
                    ub=None, # Upperbound ( x <= up), if unknown leave it as None 
                    tolerance=1e-4, # Tolerance for equality restrictions (A_eqx = b_eq)
                    device='cpu',
                    solver_precision=np.float64
                    ): 
    ## Set Device
    polytope_device = polytope.device
    polytope.send_to_device('cpu')
    
    ## Set Precision
    if '64' in str(solver_precision):
        solver_precision_tensor = torch.float64
    elif '32' in str(solver_precision):
        solver_precision_tensor = torch.float32
    elif '16' in str(solver_precision):
        solver_precision_tensor = torch.float16
    original_precision = polytope.dtype
    polytope.cast_precision(solver_precision_tensor)
    
                  
    A_in = polytope.A_in.numpy()
    b_in = polytope.b_in.numpy()
    A_in_norm = np.sqrt(np.sum(A_in**2, axis=-1))[np.newaxis,:]
    # Create new restriction matrices
    A_in_norm = np.concatenate((A_in, A_in_norm.transpose()), axis=1)
    c = np.concatenate((np.zeros(A_in.shape[1]), [-1.]))
    del A_in 
    gc.collect()
    
    ## Inequality
    if isinstance(polytope,NFDPolytope):
        mE = polytope.mE
        A_eq = polytope.A_eq.numpy()
        # The equality restrictions have zero norm. Is transposed to keep order
        A_eq_norm = np.zeros((1, A_eq.shape[0]))
        # Create new restriction matrices
        A_eq_norm = np.concatenate((A_eq, A_eq_norm.transpose()), axis=1)
        del A_eq
        gc.collect()
        b_eq = polytope.b_eq.numpy()
    else:
        mE=0
        A_eq_norm = None
        b_eq = None
        
    r = linprog(c=c,
        A_ub=A_in_norm,
        b_ub=b_in,
        A_eq=A_eq_norm,
        b_eq=b_eq,
        bounds = (lb, ub),
        method= 'highs')

    status = {0:'Optimization proceeding nominally.',
                1: 'Iteration limit reached.',
                2: 'Problem appears to be infeasible.',
                3: ' Problem appears to be unbounded.',
                4: ' Numerical difficulties encountered.',
            }[r.status]
    
    print('\nSimplex Status for the Chebyshev Center\n', status)
    
    if ('cuda' not in device) & ('16' in str(original_precision)):
        warnings.warn('Float16 precision was chosen for the polytope, but the "device=cpu" option is selected. Tensors will be temporarily cast to float32 for stability evaluation. If you wish to use float16 precision, please select "device=cuda".')
        dtype = torch.float32
    else:
        dtype = polytope.dtype
    
    x0 = torch.tensor(np.array(r.x[:-1], ndmin=2).transpose(), dtype=dtype).to(device)
    x0_in = torch.matmul(polytope.A_in.to(device).to(dtype), x0.to(device).to(dtype))
    b_in = torch.from_numpy(b_in).to(device).to(dtype)
    
    
    assert(torch.all(x0_in <= b_in )), f'Point {x0} does not satisfy A_inx <= b_in restrictions, it may have numerical inestability'
    if mE>0:
        x0_eq = torch.matmul(polytope.A_eq.to(device).to(dtype), x0.to(device).to(dtype))
        b_eq = torch.from_numpy(b_eq).to(device).to(dtype)
        assert(torch.all(torch.abs(x0_eq - b_eq) <= tolerance)), f'Point {x0} does not satisfy A_eqx = b_eq restrictions with tolerance {tolerance}, it may have numerical inestability. Try improving precision in the polytope and solver.'
    
    polytope.send_to_device(polytope_device)    
    polytope.cast_precision(original_precision)
              
    
    return x0.to(polytope_device).to(original_precision)

